/*
 * Copyright 2011 maru project.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.maru.dog.bind;

import java.lang.reflect.Field;
import java.lang.reflect.Member;
import java.lang.reflect.Method;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.maru.common.Key;
import org.maru.common.KeyGen;
import org.maru.dog.core.AbstractMarkedPointBuilder;
import org.maru.dog.core.Definition;
import org.maru.dog.core.MarkedPoint;
import org.maru.dog.core.Configuration;

import static org.maru.common.util.ConditionUtil.isNotNull;

/**
 * builder class of {@link InstanceBinderImpl}.
 * This class is not thread safe.
 */
public final class InstanceBinderBuilder<T, K> extends AbstractBinderBuilder<T, K> {

    private final List<Configuration> configurations;

    @SuppressWarnings("unchecked")
    public InstanceBinderBuilder(List<Configuration> configurations, T target, K... inputs) {
        super(new InstanceBinderImpl<T, K>(KeyGen.getKey(target.getClass()), (Class<T>) target.getClass()),
                target, inputs);
        this.configurations = configurations;
    }

    @Override
    @SuppressWarnings("unchecked")
    protected Definition<K> createAndStoreInputDefinition(Key<?> key, Class<K> inputClass) {
        if (!BindingDefinitionCache.containsInputDefinition(key)) {
            AbstractMarkedPointBuilder<K> builder = new InputMarkedPointBuilder<K>(inputClass);
            builder.build();

            BindingDefinitionCache.applyInputDefinition(key, builder.getDefinition());
            return builder.getDefinition();
        } else {
            Definition<K> definition = (Definition<K>) BindingDefinitionCache.getInputDefinition(key);
            return definition;
        }
    }

    @SuppressWarnings("unchecked")
    @Override
    protected Definition<T> createAndStoreTargetDefinition(Key<?> key, Class<T> targetType) {
        Definition<T> definition = null;
        if (!BindingDefinitionCache.containsTargetDefinition(key)) {
            AbstractMarkedPointBuilder<T> builder = new TargetMarkedPointBuilder<T>(targetType);
            builder.build();
            definition = builder.getDefinition();

            BindingDefinitionCache.applyTargetDefinition(key, definition);
        } else {
            definition =  (Definition<T>) BindingDefinitionCache.getTargetDefinition(key);
        }
        return definition;
    }

    /**
     * Creates {@link InstanceBinderImpl}
     *
     * @param targetDefinition
     * @param inputDefinitions
     */
    @Override
    protected void validateAndCreateInstanceBinder(Definition<T> targetDefinition, Map<Key<?>,Definition<K>> inputDefinitions) {

        Key<?> targetTypeKey = KeyGen.getKey(targetDefinition.getType());
        for (Entry<Key<?>, MarkedPoint<T>> entry : ((BindingDefinition<T>)targetDefinition).getMarkedPoints().entrySet()) {
            Key<?> targetBoundKey = entry.getKey();
            TargetMarkedPoint<T> targetMarkedPoint = (TargetMarkedPoint<T>)entry.getValue();
            Class<?> inputClass = null;

            //  get binding configuration if it's defined.
            BindingConfiguration bindingConfiguration = searchBindingConfiguration(targetTypeKey, targetBoundKey);
            if (isNotNull(bindingConfiguration)) {
                inputClass = bindingConfiguration.getInputClass();
                Key<?> bindKey = isNotNull(bindingConfiguration.getBindName()) ? KeyGen.getKey(bindingConfiguration.getBindName()) : targetBoundKey;

                // if binding context is created, then go to next loop
                BindingContext<T, K> context = createBindingContext(targetMarkedPoint, bindKey, inputDefinitions, inputClass, bindingConfiguration);

                if (isNotNull(context)) {
                    MemberBinder<T, K> memberBinder = createMemberBinder(context);
                    instanceBinder.addMemberBinder(memberBinder);
                }
                continue;
            }

            inputClass = targetMarkedPoint.getInputClass();

            BindingContext<T, K> context = createBindingContext(targetMarkedPoint, targetBoundKey, inputDefinitions, inputClass, null);
            if (isNotNull(context)) {
                MemberBinder<T, K> memberBinder = createMemberBinder(context);
                instanceBinder.addMemberBinder(memberBinder);
            }

        }
    }

    /**
     * search and get defined {@link BindingConfiguration} if it's defined in LOCAL_CONFIGURATION_INFOS
     *
     * @param targetTypeKey
     * @param targetBoundKey
     * @return {@link BindingConfiguration}
     */
    private BindingConfiguration searchBindingConfiguration(Key<?> targetTypeKey, Key<?> targetBoundKey) {

        BindingConfiguration configuration = null;
        if (isNotNull(configurations) && configurations.size() > 0) {
            for (Configuration c : configurations) {
                Key<?> configKey = KeyGen.getKey(((BindingConfiguration)c).getBoundName());
                if (targetBoundKey.equals(configKey)) {
                    configuration = (BindingConfiguration)c;
                    break;
                }
            }
        }
        return configuration;
    }

}
